/*

*/
import scene.*;
import scene.models.*;
import pipeline.*;
import framebuffer.*;

import java.io.File;
import java.awt.Color;
import java.awt.event.*;
import java.util.ArrayList;

/**

*/
public class InteractiveModels_v2
{
   private static ArrayList<Model> modelArray = new ArrayList<Model>();

   public static void main(String[] args)
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the digit keys '0' - '9' to choose a model.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the model along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the model.");
      System.out.println("Use the 'c' key to change the solid model colors.");
      System.out.println("Use the 'e' key to change the random edge colors.");

      // Create several Model objects.
      modelArray.add( new ObjSimpleModel(new File("assets/apple.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cow.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/galleon.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/teapot.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cessna.obj")) );
      modelArray.add( new SphereModel(30, 30, 1) );
      modelArray.add( new CylinderModel(20, 20, 0.5, 1) );
      modelArray.add( new TorusModel(25, 25, 0.75, 0.25) );
      modelArray.add( new ObjSimpleModel(new File("assets/small_rhombicosidodecahedron.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/great_rhombicosidodecahedron.obj")) );

      // Push the models away from where the camera is.
      for (Model m : modelArray)
      {
         for (LineSegment ls : m.lineSegmentList)
         {
            ls.v[0].z -= 2;
            ls.v[1].z -= 2;
         }
      }

      // Give the models random colors.
      for (Model m : modelArray)
      {
         m.setColorRandom();
      }


      // Create the Scene object that we shall render.
      final Scene scene = new Scene();

      // Add two models to the Scene.
      scene.addModel(modelArray.get(1),
                     modelArray.get(modelArray.size() - 1));


      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      final InteractiveFrame app = new InteractiveFrame("Renderer 3", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );
            char c = e.getKeyChar();
            if ('d' == c)
            {
             //Pipeline.debug = ! Pipeline.debug;
               Pipeline.debug_visible = ! Pipeline.debug_visible;
             //Rasterize.debug = ! Rasterize.debug;
            }
            else if ( Character.isDigit(c) )
            {
               int n = Integer.parseInt( Character.toString(c) );
               if (n < modelArray.size() - 1)
               {
                  scene.modelList.set(0, modelArray.get(n));
               }
            }
            else if ('c' == c)
            {
               // Change the solid random color of each model.
               for (Model m : scene.modelList)
                     m.setColorRandom();
            }
            else if ('e' == c)
            {
               // Change the random color at each edge.
               for (Model m : scene.modelList)
                  for (LineSegment ls : m.lineSegmentList)
                     ls.setColorRandom();
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('s' == c) // Scale the model 10% smaller.
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].x *= 0.9;  // this is NOT correct
                  ls.v[0].y *= 0.9;
                  ls.v[0].z *= 0.9;
                  ls.v[1].x *= 0.9;
                  ls.v[1].y *= 0.9;
                  ls.v[1].z *= 0.9;
               }
            }
            else if ('S' == c) // Scale the model 10% larger.
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].x *= 1.1;  // this is NOT correct
                  ls.v[0].y *= 1.1;
                  ls.v[0].z *= 1.1;
                  ls.v[1].x *= 1.1;
                  ls.v[1].y *= 1.1;
                  ls.v[1].z *= 1.1;
               }
            }
            else if ('x' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].x += 0.1;
                  ls.v[1].x += 0.1;
               }
            }
            else if ('X' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].x -= 0.1;
                  ls.v[1].x -= 0.1;
               }
            }
            else if ('y' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].y += 0.1;
                  ls.v[1].y += 0.1;
               }
            }
            else if ('Y' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].y -= 0.1;
                  ls.v[1].y -= 0.1;
               }
            }
            else if ('z' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].z += 0.1;
                  ls.v[1].z += 0.1;
               }
            }
            else if ('Z' == c)
            {
               for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
               {
                  ls.v[0].z -= 0.1;
                  ls.v[1].z -= 0.1;
               }
            }
            // Render again.
            this.getFrameBuffer().clearFB(Color.darkGray);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );
            java.awt.Container c = (java.awt.Container)e.getSource();
            java.awt.Insets insets = c.getInsets();
            int w = c.getWidth()  - insets.left - insets.right;
            int h = c.getHeight() - insets.top  - insets.bottom;

            this.createFrameBuffer(w, h);
            this.getFrameBuffer().clearFB(Color.darkGray);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }
      };

      app.addKeyListener(app);
      app.addComponentListener(app);
   }
}
