/*

*/
package pipeline;
import scene.*;
import framebuffer.*;

/**
    This renderer takes as its input a Scene data structure and
    a FrameBuffer data structure. This renderer mutates the
    FrameBuffer's current viewport so that it is filled in with
    the rendered image of the scene represented by the Scene object.
<p>
    This implements our third rendering pipeline. It adds a fourth
    stage, clipping, between the projection and viewport stages.
    The clipping stage clips off the parts of a line segment that
    extend out of the view rectangle (within the view plane).
*/
public class Pipeline
{
   public static boolean debug = false;
   public static boolean debug_visible = false;

   /**
      Mutate the FrameBuffer's current viewport so that it holds
      the rendered image of the Scene object.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug || debug_visible )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer without any debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Model in the Scene, render
      // every LineSgement in the Model.
      for (Model model : scene.modelList)
      {
         for (LineSegment ls : model.lineSegmentList)
         {
            // Make a deep copy of the line segment.
            LineSegment ls2 = new LineSegment(ls);

            // 1. Apply the projection transformation.
            Project.project(ls2, scene.camera);

            // 2. Clip the line segment to the standard view rectangle.
            if ( Clip.clip(ls2) )
            {
               // 3. Apply the viewport transformation.
               Viewport.viewport(ls2, fb);

               // 4. Rasterize the line segment into pixels.
               Rasterize.rasterize(ls2, fb);
            }
         }
      }
   }



   /**
      Render with quite a bit of debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Model in the Scene, render
      // every LineSgement in the Model.
      for (Model model : scene.modelList)
      {
         System.out.println("=== Render Model ===================================");

         for (LineSegment ls : model.lineSegmentList)
         {
            StringBuilder result = new StringBuilder();

            // Make a deep copy of the line segment.
            LineSegment ls2 = new LineSegment(ls);

            // log interesting information to standard output
            logLineSegment("0. Input", ls2, result);

            // 1. Apply the projection transformation.
            Project.project(ls2, scene.camera);

            // log interesting information to standard output
            logLineSegment("1. Projected", ls2, result);

            // 2. Clip the line segment to the standard view rectangle.
            if ( Clip.clip(ls2) )
            {
               // log interesting information to standard output
               logLineSegment("2. Clipped", ls2, result);

               // 3. Apply the viewport transformation.
               Viewport.viewport(ls2, fb);

               // log interesting information to standard output
               logLineSegment("3. Viewport", ls2, result);

               // 4. Rasterize the line segment into pixels.
               Rasterize.rasterize(ls2, fb);

               if ( debug_visible ) System.out.println( result );
            }
            if ( ! debug_visible ) System.out.println();
         }
         System.out.println("=== End Model ===================================");
      }
      System.out.println("=== End Scene ===================================");
   }


   private static void logLineSegment(String stage, LineSegment ls, StringBuilder accumulator)
   {
      if ( debug_visible )
         accumulator.append( stage + " " + ls.toString() );
      else
         System.out.print( stage + " " + ls.toString() );
   }
}
