/*

*/
package scene;

import java.util.Random;
import java.awt.Color;

/**
    A Vertex object has three doubles which represent the coordinates
    of a point in 3-dimensional space.
<p>
    When a Vertex object is created in a client program, before the
    Vertex object moves down the graphics rendering pipeline, the
    coordinates in the Vertex will be in "camera coordinates".
<p>
    As a Vertex object moves down the graphics rendering pipeline,
    the coordinates in the Vertex will be transformed from one
    coordinate system to another.
<p>
    A vertex object also has a color (or "shade"). The color of a
    line segment is interpolated from the colors at its two endpoints.
*/
public class Vertex
{
   public double x, y, z;
   public double r = 1.0, g = 1.0, b = 1.0;  // color at this vertex


   /**
      Construct a default vertex.
   */
   public Vertex()
   {
      set(0.0, 0.0, 0.0);
   }


   /**
      Construct a new vertex using the given x, y, and z coordinates.

      @param x x-coordinate of the new Vertex
      @param y y-coordinate of the new Vertex
      @param z z-coordinate of the new Vertex
   */
   public Vertex(double x, double y, double z)
   {
      set(x, y, z);
   }


   /**
      Construct a new vertex with the given coordinates
      and the given color.

      @param x x-coordinate of the new Vertex
      @param y y-coordinate of the new Vertex
      @param z z-coordinate of the new Vertex
      @param c Color of the new Vertex
   */
   public Vertex(double x, double y, double z, Color c)
   {
      set(x, y, z);
      setColor(c);
   }


   /**
      Construct a new vertex that is a copy of another vertex.

      @param v Vertex to make a copy of
   */
   public Vertex(Vertex v)
   {
      set(v.x, v.y, v.z);
      setColor(v);
   }


   /**
      Set the coordinates of this vertex.

      @param x new x-coordinate for this Vertex
      @param y new y-coordinate for this Vertex
      @param z new z-coordinate for this Vertex
   */
   public void set(double x, double y, double z)
   {
      this.x = x;
      this.y = y;
      this.z = z;
   }


   /**
      Get the color of this vertex.

      @return the color of this Vertex object
   */
   public Color getColor()
   {
      return new Color((float)this.r, (float)this.g, (float)this.b);
   }


   /**
      Set the color of this vertex using doubles between 0 and 1.

      @param r red color value for this Vertex as a double between 0 and 1
      @param g green color value for this Vertex as a double between 0 and 1
      @param b blue color value for this Vertex as a double between 0 and 1
   */
   public void setColor(double r, double g, double b)
   {
      this.r = r;
      this.g = g;
      this.b = b;

      if ( (r < 0.0) || (r > 1.0)
        || (g < 0.0) || (g > 1.0)
        || (b < 0.0) || (b > 1.0) )
      {
         System.err.println("ERROR! Invalid double color for vertex");
         System.err.println( this.toString() );
         System.err.printf("<r,g,b> = <% .5f  % .5f  % .5f>\n", r, g, b);
         System.exit(-1);
      }
   }


   /**
      Set the color of this vertex using ints between 0 and 255.

      @param r red color value for this Vertex as an integer between 0 and 255
      @param g green color value for this Vertex as an integer between 0 and 255
      @param b blue color value for this Vertex as an integer between 0 and 255
   */
   public void setColor(int r, int g, int b)
   {
      this.r = ((float)r)/(float)255;
      this.g = ((float)g)/(float)255;
      this.b = ((float)b)/(float)255;

      if ( (r < 0) || (r > 255)
        || (g < 0) || (g > 255)
        || (b < 0) || (b > 255))
      {
         System.err.println("ERROR! Invalid int color for vertex");
         System.err.println( this.toString() );
         System.err.printf("<r,g,b> = <%d  %d  %d>\n", r, g, b);
         System.exit(-1);
      }
   }


   /**
      Set the color of this vertex using a Color object.

      @param c Color for this Vertex object
   */
   public void setColor(Color c)
   {
      setColor(c.getRed(), c.getGreen(), c.getBlue());
   }


   /**
      Set the color of this vertex using the colors from another vertex.

      @param v Vertex object to get color values from
   */
   public void setColor(Vertex v)
   {
      // copy the color from v to this vertex
      setColor(v.r, v.g, v.b);
   }


   /**
      Set the color of this vertex to a random color.
   */
   public void setColorRandom()
   {
      Random generator = new Random();
      double r = generator.nextDouble();
      double g = generator.nextDouble();
      double b = generator.nextDouble();
      setColor(r, g, b);
   }


   /**
      For debugging.

      @return String representation of this Vertex object
   */
   public String toString()
   {
      // Here is one way to get programmable precision and width.
      int p = 5;     // the precision for the following format string
      int t = p + 4; // the width for the following format string
      String format =      "(x,y,z) = (% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
      return String.format( format, x, y, z);
    //return String.format("(x,y,z) = (% .5f  % .5f  % .5f)", x, y, z);
   }
}
